package function.builtin.special;

import function.*;
import function.builtin.EVAL;
import sexpression.*;
import table.*;

public class LET extends LispFunction {

    private ArgumentValidator argumentValidator;
    private ExecutionContext executionContext;
    private SymbolTable localScope;

    public LET() {
        this.argumentValidator = new ArgumentValidator("LET");
        this.argumentValidator.setMinimumNumberOfArguments(1);
        this.argumentValidator.setFirstArgumentExpectedType(Cons.class);
        this.executionContext = ExecutionContext.getInstance();
    }

    public SExpression call(Cons argumentList) {
        argumentValidator.validate(argumentList);
        Cons variableDefinitions = (Cons) argumentList.getCar();
        Cons body = (Cons) argumentList.getCdr();

        return evaluateInScope(variableDefinitions, body);
    }

    private SExpression evaluateInScope(Cons variableDefinitions, Cons body) {
        createLocalScope(variableDefinitions);
        SExpression lastEvaluation = evaluateBody(body);
        restorePreviousScope();

        return lastEvaluation;
    }

    private SymbolTable createLocalScope(Cons variableDefinitions) {
        localScope = new SymbolTable(executionContext.getScope());
        addVariablesToScope(localScope, variableDefinitions);
        executionContext.setScope(localScope);

        return localScope;
    }

    private void addVariablesToScope(SymbolTable scope, Cons variableDefinitions) {
        validateAllPairsAreLists(variableDefinitions);

        for (; variableDefinitions.consp(); variableDefinitions = (Cons) variableDefinitions.getCdr()) {
            Cons symbolValuePair = (Cons) variableDefinitions.getCar();

            validatePair(symbolValuePair);

            Cons restOfPair = (Cons) symbolValuePair.getCdr();
            SExpression symbol = symbolValuePair.getCar();
            SExpression value = restOfPair.getCar();

            scope.put(symbol.toString(), EVAL.eval(value));
        }
    }

    private void validateAllPairsAreLists(Cons variableDefinitions) {
        ArgumentValidator variableDefinitionListValidator = new ArgumentValidator("LET|argumentList|");
        variableDefinitionListValidator.setEveryArgumentExpectedType(Cons.class);
        variableDefinitionListValidator.validate(variableDefinitions);
    }

    private void validatePair(Cons symbolValuePair) {
        ArgumentValidator pairValidator = new ArgumentValidator("LET|pair|");
        pairValidator.setExactNumberOfArguments(2);
        pairValidator.setFirstArgumentExpectedType(Symbol.class);
        pairValidator.validate((Cons) symbolValuePair);
    }

    private SExpression evaluateBody(Cons body) {
        SExpression lastEvaluation = Nil.getInstance();

        for (; body.consp(); body = (Cons) body.getCdr())
            lastEvaluation = EVAL.eval(body.getCar());

        return lastEvaluation;
    }

    private void restorePreviousScope() {
        executionContext.setScope(localScope.getParent());
    }

    public boolean evaluateArguments() {
        return false;
    }

}
